\name{gsl.make.kernel.mdsfun}
\alias{gsl.make.kernel.mdsfun}

\title{Make minimum density similarity function based on Gaussian kernel
  density estimate}

\description{
Produces a function mdsfun(i, j) with arguments 1 <= i, j <= nrow(X).
  Let p(i, j) be the minimum of the spherical Gaussian kernel density
  estimate with bandwidth h over the edge (i, j). The value of mdsfun(i,
  j) is an estimate of p(i, j) if edge (i, j) is either in the Euclidean
  minimal spanning tree or in the k-near-neighbor graph, and 0
  otherwise. The minimum p(i, j) is approximated using grid search with
  ngrid grid points. Smaller values of k speed up the
  function. Experiments indicate that going beyond k = 20 does not
  signficantly improve clustering performance.
}

\usage{
  gsl.make.kernel.mdsfun(X, bandwidth, kmax = 20, ngrid = 10,
  interactive = F) }

\arguments{
  \item{X}{sphered data matrix. Rows correspond to observations, columns
    to variables.}
  \item{bandwidth}{bandwidth of Gaussian kernel density estimate.}
  \item{kmax}{number of nearest neighbors in k-near-neighbor graph.}
  \item{ngrid}{number of grid points used in grid search for minimum
    density along an edge.}
  \item{interactive}{if T, function will print progress reports for the
    impatient.}  
}

\value{Minimum density similarity function}

\references{Stuetzle, W. and Nugent, R. (2009.) A Generalized Single
  Method for Estimating the Cluster Tree of a Density. Journal of
  Computational  and Graphical Statistics (to appear).}

\author{Werner Stuetzle}

\seealso{gsl.make.nn.mdsfun}

\examples{
  library(gslclust)
  data(iris)
  attach(iris)
  X <- cbind(Sepal.Length, Sepal.Width, Petal.Length, Petal.Width)
  X.sphered <- sphere(X)
  optimal.bandwidth <- cv.search(X.sphered)$opt.smopar
  optimal.bandwidth
  ## [1] 0.3510157
  kernel.mdsfun <- gsl.make.kernel.mdsfun(X.sphered,
                     optimal.bandwidth, interactive = TRUE)
  gsl.cluster.out <- gsl.cluster(X.sphered, kernel.mdsfun)
  rem <- gsl.runt.excess.mass(gsl.cluster.out)
  round(nrow(X) * rem, 0)[1:10]
  ## [1] 27 2 2 2 2 2 1 1 1 1
  gsl.cluster.out <- gsl.cluster(X, kernel.mdsfun, pruning.criterion =
    "excess.mass", pruning.threshold = rem[1])
  observation.labels <- gsl.observation.labels(gsl.cluster.out)
  table(observation.labels, Species)
  ##               Species
  ## cluster.labels setosa versicolor virginica
  ##              2     50          0         0
  ##              3      0         50        50
}
\keyword{nonparametric}
\keyword{multivariate}
\keyword{cluster}

